#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <io.h>

#include "SDL_image.h"
#include "SDL_mixer.h"
#include "SDL_gfxPrimitives.h"
#include "SDL_rotozoom.h"

#include "actmain.h"
#include "charbmp.h"
#include "stroke.h"
#include "comstroke.h"
#include "util.h"
#include "jpstr.h"
#include "fps.h"
#include "fpscount.h"
#include "luautil.h"

#define JP_KANJI_BUFFER_SIZE 65536   /* ͑Ώۊobt@̃TCYB */
#define JP_KANA_BUFFER_SIZE  65536   /* ͑Ώۉobt@̃TCYB */
#define STRING_KANA_MAX       128    /* ͑Ώۉ̍ő咷B\蔲䂦B */
#define STRING_LINE_MAX       1024   /* ͑Ώۍs̍ő吔B */

#define RANKING_MAX   8         /* LO̕ێB */

#define TICKS_MAX    100        /* FPSvpz̒B */

#define SIN_TABLE_MAX 64        /* Sin֐le[u̒B */

#define KEY_HISTORY_MAX    256  /* L[z̒B */
#define KEY_HISTORY_MSEC  3000  /* L[ێԁi~bjB */
#define LIFE_HISTORY_MAX   256  /* Ctz̒B */
#define LIFE_HISTORY_MSEC 3000  /* Ctێԁi~bjB */

#define STROKER_MAX 32          /* oꂳLő吔B */

#define FILE_CONFIG     "config.txt"
#define FILE_RANKING    "ranking.txt"

/* Lua҂yieldĂ̂tOB */
#define LUA_YIELD_FRAME     0x00000001
#define LUA_YIELD_KEY_ENTER 0x00000002
#define LUA_YIELD_TALK      0x00000004

/* `Ă݂͂܂gĂȂ萔B */
#define PLAYER_MAX    2 /* 2ŒōĂ̂łقƂǈӖȂB */

/* ݒeBt@CɕۑB */
typedef struct{
    int roman_typing;       /* [}łH */
    int usable_chara_flag;  /* gp\LtO */
    char font_name[256];    /* gptHg */
    char audio_title[256];          /* ^CgʂBGMt@C */
    char audio_charaselect[256];    /* LZNgʂBGMt@C */

    /* LN^[̐B̐ςƂȂB */
    int stroker_count_x;    /* LN^[ZNg̉̐B */
    int stroker_count_y;    /* LN^[ZNg̏c̐B */
}TYPING_CONFIG;

/* vO̎sԁB */
typedef enum{
    GS_CONFIG,
    GS_TITLE,
    GS_CHARASELECT,
    GS_NEXTSTAGE,
    GS_EVENT,
    GS_TYPING,
    GS_SCORE,
    GS_RANKING
}TYPING_GAMESTATE;

typedef struct{
    int input_count;
    int miss_count;
    Uint32 total_msec;
}TYPING_SCORE;

typedef struct{
    time_t regist_time;
    TYPING_SCORE score;
/*    char name[16];*/  /*  */
}TYPING_RANKING;

enum SCREEN_FADE_MODE{FADENONE=1, FADEIN, FADEOUT};

/* Q[̐is󋵁B */
typedef struct{
    int title_menu_index;
    int story_num;
    int stage_num;
    /* 񕪂̑䎌BsB */
    char talking_string[1024];
    /* 䎌̕\ԁB */
    int talking_index_to;
    /* ʂ̕\ԁB */
    int screen_shade;   /* 0-255 */
    enum SCREEN_FADE_MODE screen_fade_mode;
    /* Lua֌WB */
    lua_State *L;
    int lua_yield_state;
    int event_skip;
    int is_ending;
    /* ^CsOJňo߃~bB */
    Uint32 stage_first_tick;
    /* ֌WB */
    Mix_Music *mix_music;
    Mix_Chunk *mix_chunk_hit;
    Mix_Chunk *mix_chunk_miss;
    int mix_volume;
    /* ΐ̏ԁB */
    Uint32 battle_first_tick;  /* JEg_E */
    int battle_end;         /* tO */
    /* vC[̑SXe[WʂẴ^CsOvlB */
    TYPING_SCORE score[STROKER_MAX];
    /* LOBŌ̂PȊOt@CɕۑB */
    TYPING_RANKING typing_ranking[RANKING_MAX+1];
}STAGE_STATE;

/* Ōۑp̍\́Bߋ莞Ԃ̑Ōx߂̂ɎgpB */
typedef struct{
    Uint32 tick;    /* Ōꂽ̌o߃~b */
    int hit;        /* Ō */
    int miss;       /* ~XŌ */
}KEY_HISTORY;

/* Ctۑp̍\́Bߋ莞Ԃ̃Ctω߂̂ɎgpB */
typedef struct{
    Uint32 tick;    /* Ctω̌o߃~b */
    int life_plus;  /* Ctω */
}LIFE_HISTORY;

/* Ō҈l̏ԁB */
typedef struct{
    int checked_count;          /* r܂ŃqbgL[͂̕B */
    char input_string[256];     /* ͂L[͕iʕ\pj */
    char input_stringcode[256]; /* ͂L[͕i`FbNpj */
    int input_count;            /* ͂L[͂̕B */
    size_t jp_input_count;      /* ͂{̃oCgB */
    int jp_current_line;        /* ̍sCfbNXB */

    /* COM̏ԁBNULLȂvC[B */
    COMSTROKER_STATUS *com_stroker;

    /* ^CsOvlBPXe[WB */
    TYPING_SCORE score;

    /* 牺̓^CsOɒڊ֌WȂϐB */
    int life;                   /* Ct */
    int life_gage;              /* ʂɕ\郉CtiXɕωj */
    KEY_HISTORY key_history[KEY_HISTORY_MAX];       /* Ō */
    int key_history_index;
    LIFE_HISTORY life_history[LIFE_HISTORY_MAX];    /* Ctω */
    int life_history_index;
    int chara_index;            /* LN^[ԍ */
/*    int player_index;*/
    int animation_count;
    int shade_rate;
    int shade_rate_to;
}STROKER_STATUS;

/* ͑ΏەB */
typedef struct{
    char jp_buffer_kanji[JP_KANJI_BUFFER_SIZE]; /* iɕ\pj */
    char *jp_string_kanji[STRING_LINE_MAX];
    char jp_buffer_kana[JP_KANA_BUFFER_SIZE];   /* Ђ炪ȁi\͔pj */
    char *jp_string_kana[STRING_LINE_MAX];
    int jp_string_count;                        /* sB */
    /* fobOpIvV */
    int noswap_option;  /* ͕XbvB */
    int allcom_option;  /* COM */
}JPSTRING_DATA;

/* Q[ԑSB */
typedef struct{
    /* o߃~bB */
    Uint32 last_tick;
    /* Jñt[B */
    int frame;
    /* FPS\tOB */
    int fps_show;
    /* FPSZopzB */
    FPS_COUNTER *fps_counter;
    /* ݒEŜ̏ԂȂǁB */
    TYPING_CONFIG config;
    TYPING_GAMESTATE gamestate;
    STAGE_STATE stage;
    JPSTRING_DATA jpstrdata;
    /* Ō҂̏ԁBɂQlΐԁB */
    STROKER_STATUS strokers[PLAYER_MAX];
}ALL_GAME_STATUS;

/* `pT[tFCXȂǁB̍\̂ɓꂽƂ낾ʓ|Ă߂B */
static SDL_Surface *surface_stroker[STROKER_MAX][2];
static SDL_Surface *surface_stroker_shade[STROKER_MAX][2];
static SDL_Surface *surface_stroker_name[STROKER_MAX];
static SDL_Surface *surface_stroker_face[STROKER_MAX+1];  /* Ō̂P̓nei̕B */
static SDL_Surface *surface_title;
static SDL_Surface *surface_bg;
static SDL_Surface *surface_event_bg;
static SDL_Surface *surface_gage;
static SDL_Surface *surface_winlose;
/* \pT[tFCXǗ\́B */
static CHR_FONT_BMP font_bmp;   /* e[uS */
static CHR_FONT_BMP font_num96; /*  96pt */

/* Sin֐le[uB256{lB~B */
static int sin_table[SIN_TABLE_MAX];

/* LuasɗLȒlBȊO̎NULLB */
/* ̓O[oϐɂ͂ȂKvȎlƂƂőËB */
static ALL_GAME_STATUS *all_game_status;

/*<---------------------------------------------------------------------->*/
/* iIݒǂݍށB */
static void load_config(TYPING_CONFIG *config)
{
    FILE *fp;
    char buf[256];
    char *p;

    /* ƂȂftHglݒBɈӖȂB */
    config->roman_typing = 1;
    config->usable_chara_flag = 1;
    config->stroker_count_x = 3;
    config->stroker_count_y = 2;

    fp = fopen(FILE_CONFIG, "r");
    if(fp == NULL){
        return;
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        sscanf(buf, "%d", &config->roman_typing);
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        sscanf(buf, "%d", &config->usable_chara_flag);
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        strcpy(config->font_name, buf);
        p = strchr(config->font_name, '\n');
        if(p != NULL)
            *p = '\0';
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        sscanf(buf, "%d", &config->stroker_count_x);
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        sscanf(buf, "%d", &config->stroker_count_y);
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        strcpy(config->audio_title, buf);
        p = strchr(config->audio_title, '\n');
        if(p != NULL)
            *p = '\0';
    }
    if(fgets(buf, sizeof buf, fp) != NULL){
        strcpy(config->audio_charaselect, buf);
        p = strchr(config->audio_charaselect, '\n');
        if(p != NULL)
            *p = '\0';
    }
    fclose(fp);
}
/*<---------------------------------------------------------------------->*/
/* iIݒށB */
static void save_config(const TYPING_CONFIG *config)
{
    FILE *fp;

    fp = fopen(FILE_CONFIG, "w");
    if(fp == NULL){
        return;
    }
    fprintf(fp, "%d\n", config->roman_typing);
    fprintf(fp, "%d\n", config->usable_chara_flag);
    fprintf(fp, "%s\n", config->font_name);
    fprintf(fp, "%d\n", config->stroker_count_x);
    fprintf(fp, "%d\n", config->stroker_count_y);
    fprintf(fp, "%s\n", config->audio_title);
    fprintf(fp, "%s\n", config->audio_charaselect);
    fclose(fp);
}
/*<---------------------------------------------------------------------->*/
/* LOǂݍށB */
static void load_ranking(TYPING_RANKING *ranking, int ranking_count)
{
    FILE *fp;
    char buf[64];
    long l;
    int i;
    TYPING_SCORE *score;

    memset(ranking, 0, sizeof(TYPING_RANKING)*ranking_count);

    fp = fopen(FILE_RANKING, "r");
    if(fp == NULL){
        return;
    }
    for(i=0;i<ranking_count;i++){
        if(fgets(buf, sizeof buf, fp) != NULL){
            score = &ranking[i].score;
            if(sscanf(buf, "%ld,%d,%d,%u", &l, &score->input_count,
                      &score->miss_count, &score->total_msec) != 4){
                break;
            }
            ranking[i].regist_time = (time_t)l;
        }else{
            break;
        }
    }
    fclose(fp);
}
/*<---------------------------------------------------------------------->*/
/* LOށB */
static void save_ranking(const TYPING_RANKING *ranking, int ranking_count)
{
    FILE *fp;
    int i;
    const TYPING_SCORE *score;

    fp = fopen(FILE_RANKING, "w");
    if(fp == NULL){
        return;
    }
    for(i=0;i<ranking_count;i++){
        score = &ranking[i].score;
        fprintf(fp, "%ld,%d,%d,%u\n", (long)ranking[i].regist_time,
                score->input_count, score->miss_count, score->total_msec);
    }
    fclose(fp);
}
/*<---------------------------------------------------------------------->*/
/* ͑ΏۂƂȂ{ꕶǂݍށB */
static int load_jp_string(const char *file, JPSTRING_DATA *jpstrdata)
{
    FILE *fp;
    char *p;
    char *kanji_buf, *kana_buf; /* ǂݍ݈ʒuւ̃|C^B */
    int kanji_rest, kana_rest;  /* obt@̎ceʁB */
    int line_size;
    int line = 0;
    int i,n;                    /* Xbvp */

    fp = fopen(file, "r");
    if(fp == NULL){
        return -1;
    }
    kanji_buf = jpstrdata->jp_buffer_kanji;
    kana_buf = jpstrdata->jp_buffer_kana;
    kanji_rest = JP_KANJI_BUFFER_SIZE;
    kana_rest = JP_KANA_BUFFER_SIZE;
    jpstrdata->jp_string_count = 0;
    for(;;){
        if(jpstrdata->jp_string_count >= STRING_LINE_MAX){
            fprintf(stderr, "obt@ӂ܂Bő%ds܂łłB\n", STRING_LINE_MAX);
            break;
        }
        /* ssǂݍށB */
        p = fgets(kanji_buf, kanji_rest, fp);
        if(p == NULL){
            break;
        }
        line++;
        p = strchr(kanji_buf, '\n');
        if(p != NULL){
            *p = '\0';
        }
        jpstrdata->jp_string_kanji[jpstrdata->jp_string_count] = kanji_buf;
        kanji_buf = strchr(kanji_buf, '\0') + 1;
        line_size = kanji_buf - jpstrdata->jp_string_kanji[jpstrdata->jp_string_count];
        kanji_rest -= line_size;
        /* Ђ炪ȍssǂݍށB */
        p = fgets(kana_buf, kana_rest, fp);
        if(p == NULL){
            break;
        }
        line++;
        p = strchr(kana_buf, '\n');
        if(p != NULL){
            *p = '\0';
        }
        jpstrdata->jp_string_kana[jpstrdata->jp_string_count] = kana_buf;
        kana_buf = strchr(kana_buf, '\0') + 1;
        line_size = kana_buf - jpstrdata->jp_string_kana[jpstrdata->jp_string_count];
        kana_rest -= line_size;
        if(line_size > STRING_KANA_MAX){
            fprintf(stderr, "%dsɒ镶񂪂܂Bő%doCg܂łłB\n", line, STRING_KANA_MAX);
            break;
        }
        jpstrdata->jp_string_count++;
    }
    /* XbvIvVw肳ĂȂ΁A_ɕёւB */
    if(!jpstrdata->noswap_option){
        for(i=0; i<jpstrdata->jp_string_count; i++){
            /* swap֐KvB */
            n = rand() / (RAND_MAX / jpstrdata->jp_string_count + 1);
            p = jpstrdata->jp_string_kanji[i];
            jpstrdata->jp_string_kanji[i] = jpstrdata->jp_string_kanji[n];
            jpstrdata->jp_string_kanji[n] = p;
            p = jpstrdata->jp_string_kana[i];
            jpstrdata->jp_string_kana[i] = jpstrdata->jp_string_kana[n];
            jpstrdata->jp_string_kana[n] = p;
        }
    }
    fclose(fp);
#if 0
    for(i=0;i<jpstrdata->jp_string_count;i++){
        printf("%s\n", jpstrdata->jp_string_kana[i]);
    }
#endif
    return 0;
}
/*<---------------------------------------------------------------------->*/
static int load_images(const TYPING_CONFIG *config)
{
    SDL_Rect r;
    char filename[32];
    char str[16];
    int index;
    SDL_Surface *surface;

    surface_title = IMG_Load("image/title.png");
    if(surface_title == NULL){
        r.x = 0;
        r.y = 0;
        r.w = 640;
        r.h = 480;
        surface_title = SDL_CreateRGBSurface(SDL_SWSURFACE, r.w, r.h,
                32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
        SDL_FillRect(surface_bg, &r, SDL_MapRGB(surface_bg->format, 64, 128, 64));
    }
    surface_bg = IMG_Load("image/zian.png");
    if(surface_bg == NULL){
        r.x = 0;
        r.y = 0;
        r.w = 640;
        r.h = 480;
        surface_bg = SDL_CreateRGBSurface(SDL_SWSURFACE, r.w, r.h,
                32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
        SDL_FillRect(surface_bg, &r, SDL_MapRGB(surface_bg->format, 0, 0, 0));
    }
    surface_gage = IMG_Load("image/gage.png");
    if(surface_gage == NULL){
        r.x = 0;
        r.y = 0;
        r.w = 640;
        r.h = 480;
        surface_gage = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
        SDL_FillRect(surface_gage, &r, SDL_MapRGB(surface_gage->format, 0, 0, 128));
    }
    SDL_SetColorKey(surface_gage, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_gage->format, 0, 0, 128));
    surface_winlose = IMG_Load("image/winlose.png");
    if(surface_winlose == NULL){
        r.x = 0;
        r.y = 0;
        r.w = 400;
        r.h = 96;
        surface_winlose = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
        SDL_FillRect(surface_winlose, &r, SDL_MapRGB(surface_winlose->format, 0, 0, 128));
    }
    SDL_SetColorKey(surface_winlose, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_winlose->format, 0, 128, 128));

    for(index=0; index<config->stroker_count_x*config->stroker_count_y; index++){
        /* 摜̓ǂݍ݁B */
        sprintf(filename, "image/%02d.png", index);
        surface = IMG_Load(filename);
        if(surface == NULL){
    /*        fprintf(stderr, "LoadBMP Ɏs܂: %s\n", SDL_GetError());
            return -1;*/
            r.x = 0;
            r.y = 0;
            r.w = 224;
            r.h = 320;
            surface = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                    32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);
            SDL_FillRect(surface, &r, SDL_MapRGB(surface->format, 0, 0, 255));
            sprintf(str, "%d", index);
            draw_string(&font_bmp, surface, 0, 0, str);
        }
        surface_stroker[index][0] = surface;
        /* ]T[tFCX쐬B */
        surface = zoomSurface(surface_stroker[index][0], -1., 1., 0);
        if(surface_stroker[index][0]->format->Amask == 0){
            /* `lꍇAߐFݒ肷B */
            /* ʂȂ悤zoomSurface()͕K`lݒ肵Ă܂̂łނȂB */
            surface_stroker[index][1] = SDL_ConvertSurface(surface, surface_stroker[index][0]->format, SDL_SWSURFACE|SDL_SRCALPHA);
            SDL_SetColorKey(surface_stroker[index][0], SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_winlose->format, 0, 0, 128));
            SDL_SetColorKey(surface_stroker[index][1], SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_winlose->format, 0, 0, 128));
        }
        else{
            /* `lꍇÂ܂܁B */
            surface_stroker[index][1] = surface;
        }

        /* e摜̓ǂݍ݁B */
        sprintf(filename, "image/%02d_shade.png", index);
        surface = IMG_Load(filename);
        if(surface == NULL){
            r.x = 0;
            r.y = 0;
            r.w = 224;
            r.h = 320;
            surface = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                    32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
            SDL_FillRect(surface, &r, SDL_MapRGB(surface->format, 0, 0, 0));
        }
        surface_stroker_shade[index][0] = surface;
        /* ]T[tFCX쐬B */
        surface = zoomSurface(surface_stroker_shade[index][0], -1., 1., 0);
        /* ʂȂ悤zoomSurface()͕K`lݒ肵Ă܂̂łނȂB */
        surface_stroker_shade[index][1] = SDL_ConvertSurface(surface, surface_stroker_shade[index][0]->format, SDL_SWSURFACE|SDL_SRCALPHA);
        SDL_FreeSurface(surface);
        /* ߐFݒB */
        SDL_SetColorKey(surface_stroker_shade[index][0], SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_stroker_shade[index][0]->format, 0, 0, 128));
        SDL_SetColorKey(surface_stroker_shade[index][1], SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface_stroker_shade[index][1]->format, 0, 0, 128));
        /* O摜̓ǂݍ݁B */
        sprintf(filename, "image/%02d_name.png", index);
        surface = IMG_Load(filename);
        if(surface == NULL){
            r.x = 0;
            r.y = 0;
            r.w = 256;
            r.h = 32;
            surface = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                    32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
            SDL_FillRect(surface, &r, SDL_MapRGB(surface->format, 0, 128, 0));
        }
        SDL_SetColorKey(surface, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface->format, 0, 0, 128));
        surface_stroker_name[index] = surface;
    }

    for(index=0; index<config->stroker_count_x * config->stroker_count_y+1; index++){
        /* 摜̓ǂݍ݁Bneî̕ŕʃ[vB */
        sprintf(filename, "image/%02d_face.png", index);
        surface = IMG_Load(filename);
        if(surface == NULL){
            r.x = 0;
            r.y = 0;
            r.w = 64;
            r.h = 64;
            surface = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCALPHA, r.w, r.h,
                    32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000);
            SDL_FillRect(surface, &r, SDL_MapRGB(surface->format, 0, 0, 255));
            sprintf(str, "%d", index);
            draw_string(&font_bmp, surface, 0, 0, str);
        }
        SDL_SetColorKey(surface, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(surface->format, 0, 0, 128));
        surface_stroker_face[index] = surface;
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
static int load_audio_and_play(STAGE_STATE *stage, const char *filename)
{
    char path[256];

    /* BGMĐȂ~߂B */
    /* ̏͂͂ubN̂Ŏۂɂ͋NȂ悤ɑ̏Ő䂷邱ƁB */
    if(stage->mix_music != NULL){
        if(Mix_PlayingMusic()){
            Mix_FadeOutMusic(500);
            while(Mix_PlayingMusic()){
                SDL_Delay(100);
            }
        }
        Mix_FreeMusic(stage->mix_music);
    }
    sprintf(path, "sound/mid/%s", filename);
    stage->mix_music = Mix_LoadMUS(path);
    if(stage->mix_music != NULL){
        Mix_PlayMusic(stage->mix_music, -1);
    }
    else{
        fprintf(stderr, "Mix_LoadMUS error. %s\n", Mix_GetError());
        return -1;
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* RtBOʏB */
static int callback_config(const SDL_keysym *keysyms, int keysym_count, TYPING_CONFIG *config)
{
    SDLKey sym;
    int i;

    for(i=0;i<keysym_count;i++){
        sym = keysyms[i].sym;
        if(sym == SDLK_UP && !config->roman_typing){
            config->roman_typing = 1;
        }
        else if(sym == SDLK_DOWN && config->roman_typing){
            config->roman_typing = 0;
        }
        else if(sym == SDLK_RETURN){
            /* [}Jîǂ炩̐ݒǂݍށB */
            if(config->roman_typing){
                if(load_roman_charset("txt/roman.txt", 1) == -1){
                    printf("failed to load roman pattern txt.\n");
                    return -1;
                }
            }
            else{
                if(load_roman_charset("txt/kana.txt", 0) == -1){
                    printf("failed to load kana pattern txt.\n");
                    return -1;
                }
            }
            return 0;
        }
    }
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* RtBOʕ`B */
static void drawscreen_config(SDL_Surface *screen, int frame, const TYPING_CONFIG *config)
{
    Uint32 black, rcolor;
    SDL_Rect r;
    int sx,sy,sw,sh;
    int step;

    black = SDL_MapRGB(screen->format, 0, 0, 0);
    r.x = 0;
    r.y = 0;
    r.w = screen->w;
    r.h = screen->h;
    SDL_FillRect(screen, &r, black);

    sw = font_bmp.chr_w * 8;
    sh = font_bmp.chr_h;
    sx = (screen->w - sw)/2;
    sy = (screen->h - sh*3)/2;

    if(frame%60 < 30){
        step = frame%30;
    }else{
        step = 30 - frame%30;
    }
    rcolor = SDL_MapRGB(screen->format, 0, 64+step*4, 128);
    if(config->roman_typing){
        r.x = sx;
        r.y = sy;
        r.w = sw;
        r.h = sh;
        SDL_FillRect(screen, &r, rcolor);
    }else{
        r.x = sx;
        r.y = sy + sh*2;
        r.w = sw;
        r.h = sh;
        SDL_FillRect(screen, &r, rcolor);
    }

    draw_string(&font_bmp, screen, sx, sy,      "[}");
    draw_string(&font_bmp, screen, sx, sy+sh*2, "ȓ");
}
/*<---------------------------------------------------------------------->*/
/* ^CgʊJnBI͗B */
static int start_title(ALL_GAME_STATUS *status)
{
    load_audio_and_play(&status->stage, status->config.audio_title);
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* ^CgʏB */
static int callback_title(const SDL_keysym *keysyms, int keysym_count, STAGE_STATE *stage)
{
    SDLKey sym;
    int i;
    int cont = 1;

    /* tF[hCEtF[hAEg */
    if(stage->screen_fade_mode == FADEIN){
        stage->screen_shade = max(0, stage->screen_shade - 16);
        if(!stage->screen_shade){
            stage->screen_fade_mode = FADENONE;
        }
    }
    else if(stage->screen_fade_mode == FADEOUT){
        stage->screen_shade = min(255, stage->screen_shade + 16);
        if(stage->screen_shade == 255){
            stage->screen_fade_mode = FADEIN;
            cont = 0;
        }
    }

    for(i=0;i<keysym_count;i++){
        sym = keysyms[i].sym;
        if(sym == SDLK_UP){
            stage->title_menu_index--;
            if(stage->title_menu_index < 0)
                stage->title_menu_index = 2;
        }
        else if(sym == SDLK_DOWN){
            stage->title_menu_index++;
            if(stage->title_menu_index > 2)
                stage->title_menu_index = 0;
        }
        else if(sym == SDLK_RETURN){
            if(Mix_PlayingMusic()){
                Mix_FadeOutMusic(500);
            }
            stage->screen_fade_mode = FADEOUT;
        }
    }
    return cont;
}
/*<---------------------------------------------------------------------->*/
/* ^Cgʕ`B */
static void drawscreen_title(SDL_Surface *screen, const STAGE_STATE *stage, int frame)
{
    Uint32 rcolor;
    SDL_Rect srcrect, dstrect;
    int sx,sy,sw,sh;
    int mh;
    int step;

    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface_title->w;
    dstrect.h = surface_title->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface_title->w;
    srcrect.h = surface_title->h;
    SDL_BlitSurface(surface_title, &srcrect, screen, &dstrect);

    sw = font_bmp.chr_w * 8;
    sh = font_bmp.chr_h;
    sx = (screen->w - sw)/2;
    sy = screen->h/2 + 64;

    if(frame%60 < 30){
        step = frame%30;
    }else{
        step = 30 - frame%30;
    }

    mh = sh*3/2;
    boxColor(screen, sx-8, sy-8, sx+87, sy+mh*3+7, 0x00000080);
    rcolor = ((64+step*4)<<16) | 0x8000 | (135+step*4);
    boxColor(screen, sx, sy+mh*stage->title_menu_index, sx+79, sy+mh*stage->title_menu_index+sh-1, rcolor);

    draw_string(&font_bmp, screen, sx, sy,      "Xg[[");
    draw_string(&font_bmp, screen, sx, sy+mh,   "LO");
    draw_string(&font_bmp, screen, sx, sy+mh*2, "I");
}
/*<---------------------------------------------------------------------->*/
static int start_charaselect(ALL_GAME_STATUS *status)
{
    int i;

    for(i=0;i<PLAYER_MAX;i++){
        status->strokers[i].animation_count = 0;
    }
    load_audio_and_play(&status->stage, status->config.audio_charaselect);
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* LN^[ZNgʏBʓ|Ȃ̂ŃL̑Ƃln[hR[fBOB */
static int callback_charaselect(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    const TYPING_CONFIG *config;
    STROKER_STATUS *stroker;
    STAGE_STATE *stage;
    SDLKey sym;
    int i;
    int x,y;
    int index_old;
    int cont = 1;

    config = &status->config;
    stroker = &status->strokers[0];
    stage = &status->stage;
    /* tF[hCEtF[hAEg */
    if(stage->screen_fade_mode == FADEIN){
        stage->screen_shade = max(0, stage->screen_shade - 16);
        if(!stage->screen_shade){
            stage->screen_fade_mode = FADENONE;
        }
    }
    else if(stage->screen_fade_mode == FADEOUT){
        stage->screen_shade = min(255, stage->screen_shade + 16);
        if(stage->screen_shade == 255){
            stage->screen_fade_mode = FADEIN;
            cont = 0;
        }
    }

    index_old = stroker->chara_index;
    if(cont){
        x = (stroker->chara_index)%config->stroker_count_x;
        y = (stroker->chara_index)/config->stroker_count_x;
        for(i=0;i<keysym_count;i++){
            sym = keysyms[i].sym;
            if(sym == SDLK_UP){
                if(y > 0)
                    y--;
            }
            else if(sym == SDLK_DOWN){
                if(y < config->stroker_count_y-1)
                    y++;
            }
            else if(sym == SDLK_LEFT){
                if(x > 0)
                    x--;
            }
            else if(sym == SDLK_RIGHT){
                if(x < config->stroker_count_x-1)
                    x++;
            }
            else if(sym == SDLK_RETURN){
                /* I\ȃLȂAIB */
                if(config->usable_chara_flag & (1<<(stroker->chara_index))){
                    if(Mix_PlayingMusic()){
                        Mix_FadeOutMusic(500);
                    }
                    stage->screen_fade_mode = FADEOUT;
                }
            }
            stroker->chara_index = y*config->stroker_count_x + x;
        }
    }
    if(index_old != stroker->chara_index){
        stroker->animation_count = 0;
    }
    else{
        stroker->animation_count++;
    }
    return cont;
}
/*<---------------------------------------------------------------------->*/
/* G`悷BƂ肠YW͌ŒŁBrev = 0 or 1 */
static void draw_stroker_figure(SDL_Surface *screen, int x, int index, int shade, int reverse)
{
    SDL_Surface *surface;
    SDL_Rect r, dstrect, srcrect;

    surface = surface_stroker[index][reverse];
    dstrect.x = x;
    dstrect.y = 480 - surface->h;
    dstrect.w = surface->w;
    dstrect.h = surface->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface->w;
    srcrect.h = surface->h;
    r = dstrect;    /* SDL_BlitSurface ͕`̋`Ă܂̂ŁB */
    SDL_BlitSurface(surface, &srcrect, screen, &r);
    if(shade){
        surface = surface_stroker_shade[index][reverse];
        SDL_SetAlpha(surface, SDL_SRCALPHA|SDL_RLEACCEL, shade);
        SDL_BlitSurface(surface, &srcrect, screen, &dstrect);
    }
}
/*<---------------------------------------------------------------------->*/
/* 摜`悷B */
static void draw_stroker_face(SDL_Surface *screen, int x, int y, int index)
{
    SDL_Surface *surface;
    SDL_Rect dstrect, srcrect;

    surface = surface_stroker_face[index];
    dstrect.x = x;
    dstrect.y = y;
    dstrect.w = surface->w;
    dstrect.h = surface->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface->w;
    srcrect.h = surface->h;
    boxColor(screen, x, y, x+63, y+63, 0x00000080);
    SDL_BlitSurface(surface, &srcrect, screen, &dstrect);
}
/*<---------------------------------------------------------------------->*/
/* O摜`悷B */
static void draw_stroker_name(SDL_Surface *screen, int x, int y, int index)
{
    SDL_Surface *surface;
    SDL_Rect dstrect, srcrect;

    surface = surface_stroker_name[index];
    dstrect.x = x;
    dstrect.y = y;
    dstrect.w = surface->w;
    dstrect.h = surface->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface->w;
    srcrect.h = surface->h;
    SDL_BlitSurface(surface, &srcrect, screen, &dstrect);
}
/*<---------------------------------------------------------------------->*/
/* LN^[ZNgʕ`B */
static void drawscreen_charaselect(SDL_Surface *screen, int frame, const TYPING_CONFIG *config, const STROKER_STATUS *stroker, int usable_chara_flag)
{
    const int img_w=80, img_h = 80;
    int list_x, list_y;
    SDL_Rect srcrect, dstrect;
    int cx, cy;
    int x, y;
    int index;
    int slide;
    int count, count_max;
    SDL_Surface *surface;

    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface_bg->w;
    dstrect.h = surface_bg->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface_bg->w;
    srcrect.h = surface_bg->h;
    SDL_BlitSurface(surface_bg, &srcrect, screen, &dstrect);

    list_x = screen->w/2 - img_w*config->stroker_count_x/2;
    list_y = screen->h   - img_h*config->stroker_count_y - 32;

    for(y=0; y<config->stroker_count_y; y++){
        for(x=0; x<config->stroker_count_x; x++){
            index = y*config->stroker_count_x + x;
            /* ݑIs\ȃL̓neiɂB */
            if(!(usable_chara_flag & (1<<index))){
                index = config->stroker_count_x * config->stroker_count_y;      /* nei */
            }
            draw_stroker_face(screen, list_x + x*img_w, list_y + y*img_h, index);
        }
    }

    cx = stroker->chara_index%config->stroker_count_x;
    cy = stroker->chara_index/config->stroker_count_x;

    x = list_x + cx*img_w;
    y = list_y + cy*img_h;
    slide = sin_table[(frame<<2)%SIN_TABLE_MAX]>>6;
    rectangleColor(screen, x-4-slide, y-4-slide, x+67+slide, y+67+slide, 0x00808080);
    rectangleColor(screen, x-5-slide, y-5-slide, x+68+slide, y+68+slide, 0x00808080);

    if(usable_chara_flag & (1<<stroker->chara_index)){
        count_max = SIN_TABLE_MAX/4 - 1;
        if(stroker->animation_count < count_max){
            count = stroker->animation_count;
        }else{
            count = count_max;
        }
        surface = surface_stroker[stroker->chara_index][0];
        x = -surface->w + (surface->w * sin_table[count] + 255)/256;
        draw_stroker_figure(screen, x, stroker->chara_index, 255 - 255*count/count_max, 0);
        draw_stroker_name(screen, 16, 480-sin_table[count]/4, stroker->chara_index);
    }
}
/*<---------------------------------------------------------------------->*/
/* 䎌f[^ZbgB */
static int LF_talk( lua_State *L )
{
    STAGE_STATE *stage;
    const char *str;
    int n;

    n = lua_gettop(L);
    if(n != 1 || !lua_isstring(L,1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "talk");
        return 0;
    }

    str = lua_tostring(L, 1);
    stage = &all_game_status->stage;
    /* 񂪒G[B */
    if(strlen(str) >= sizeof stage->talking_string){
        lua_settop(L, 0);
        return 0;
    }
    strcpy(stage->talking_string, str);
    stage->talking_index_to = 0;
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* 摜̍Fт̔䗦ZbgB */
static int LF_figure_shade_set( lua_State *L )
{
    int player_index;
    int n;

    n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "figure_shade_set");
        return 0;
    }

    player_index = (int)lua_tonumber(L, 1);
    if(player_index < 0 || player_index >= PLAYER_MAX){
        lua_settop(L, 0);
        return 0;
    }
    all_game_status->strokers[player_index].shade_rate_to = (int)lua_tonumber(L, 2);
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* GfBOɏԑJڂB */
static int LF_set_ending( lua_State *L )
{
    all_game_status->stage.is_ending = 1;
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* gp\LtO̐ݒB */
static int LF_set_usable_chara_flag( lua_State *L )
{
    int n;

    n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "api_set_usable_chara_flag");
        return 0;
    }
    all_game_status->config.usable_chara_flag = lua_tonumber(L, 1);
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* wiw肷B */
static int LF_bg_set( lua_State *L )
{
    int n;
    char path[256];

    n = lua_gettop(L);
    if(n != 1 || !lua_isstring(L,1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "api_bg_set");
        return 0;
    }

    sprintf(path, "image/%s", lua_tostring(L, 1));
    if(surface_event_bg != NULL)
        SDL_FreeSurface(surface_event_bg);
    surface_event_bg = IMG_Load(path);
    if(surface_event_bg == NULL){
        fprintf(stderr, "failed to load image. %s\n", path);
        return 0;
    }
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* aflw肷B */
static int LF_bgm_set( lua_State *L )
{
    int n;

    n = lua_gettop(L);
    if(n != 1 || !lua_isstring(L,1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "api_bgm_set");
        return 0;
    }

    if(load_audio_and_play(&all_game_status->stage, lua_tostring(L, 1)) == -1){
        return 0;
    }
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* afltF[hAEgB */
static int LF_bgm_fadeout( lua_State *L )
{
    int n;
    int msec;

    n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L, 1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "api_bgm_fadeout");
        return 0;
    }

    msec = (int)lua_tonumber(L, 1);
    lua_settop(L, 0);
    lua_pushboolean(L, Mix_FadeOutMusic(msec));
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* ʑŜ̈ÐF䗦擾^肷B */
static int LF_fade_set( lua_State *L )
{
    int n;
    int shade;

    n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L, 1)){
        lua_settop(L, 0);
        luautil_arg_error_log(L, "api_fade_set");
        return 0;
    }

    shade = (int)lua_tonumber(L, 1);
    if(shade >= 0 && shade <= 255){
        all_game_status->stage.screen_shade = shade;
    }
    lua_settop(L, 0);
    lua_pushnumber(L, all_game_status->stage.screen_shade);
    return 1;
}
/*<---------------------------------------------------------------------->*/
/* ̃t[܂ŃXNvg̎s𒆒fB */
static int LF_wait_frame( lua_State *L )
{
/*    printf("LF_wait_frame\n");*/
    all_game_status->stage.lua_yield_state |= LUA_YIELD_FRAME;
    return lua_yield(L, 0);
}
/*<---------------------------------------------------------------------->*/
/* 䎌I҂tO𗧂ĂăXNvg̎s𒆒fB */
static int LF_wait_talk_end( lua_State *L )
{
/*    printf("LF_wait_talk_end\n");*/
    all_game_status->stage.lua_yield_state |= LUA_YIELD_TALK;
    return lua_yield(L, 0);
}
/*<---------------------------------------------------------------------->*/
/* L[͑҂tO𗧂ĂăXNvg̎s𒆒fB */
static int LF_wait_key_enter( lua_State *L )
{
/*    printf("LF_wait_key_enter\n");*/
    all_game_status->stage.lua_yield_state |= LUA_YIELD_KEY_ENTER;
    return lua_yield(L, 0);
}
/*<---------------------------------------------------------------------->*/
/* bʊJnBI͗B */
static int start_event(ALL_GAME_STATUS *status)
{
    STAGE_STATE *stage;
    STROKER_STATUS *stroker;
    int player_index;
    char funcname[32];
    int ret;

    for(player_index=0; player_index<PLAYER_MAX; player_index++){
        stroker = &status->strokers[player_index];
        stroker->shade_rate_to = stroker->shade_rate = 128;
    }

    stage = &status->stage;
    status->strokers[1].chara_index = stage->stage_num;
    stage->talking_string[0] = '\0';
    stage->talking_index_to = 0;
    stage->lua_yield_state = 0;
    stage->event_skip = 0;
    sprintf(funcname, "event_%02d_%02d", stage->story_num, stage->stage_num);
    lua_getglobal(stage->L, funcname);
    all_game_status = status;
    ret = lua_resume(stage->L, 0);
    all_game_status = NULL;
    if(ret != 0){
        fprintf(stderr, "Failed to exec event (%s)\n", funcname);
        return -1;
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* bʏB */
static int callback_event(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    STAGE_STATE *stage;
    SDLKey sym;
    STROKER_STATUS *stroker;
    const char *talking_string;
    int i, chr;
    int player_index;
    int keyhit = 0;
    int resume_flag = 0;
    int cont = 1;

    stage = &status->stage;

    if(stage->event_skip){
        /* CxgXLbvB */
        if(stage->screen_shade < 255){
            /* ܂͕ʂɃCxgsȂtF[hAEgB */
            stage->screen_shade = min(255, stage->screen_shade + 16);
        }
        else{
            /* ̂܂܂ł͎̃CxgresumełȂ̂ł܂Ƃ߂resumeB */
            all_game_status = status;
            while(!lua_resume(stage->L, 0))
                ;
            all_game_status = NULL;
            stage->screen_fade_mode = FADEIN;
            return 0;
        }
    }
    else{
        for(i=0;i<keysym_count;i++){
            sym = keysyms[i].sym;
            if(sym == SDLK_SPACE){
                keyhit = 1;
            }
            if(sym == SDLK_RETURN){
                stage->event_skip = 1;
                stage->screen_fade_mode = FADEOUT;
            }
        }
    }

    /* 摜̈ÐF䗦w肳ꂽÐF䗦ƈႤꍇAXɕωB */
    for(player_index=0; player_index<PLAYER_MAX; player_index++){
        stroker = &status->strokers[player_index];
        i = stroker->shade_rate_to - stroker->shade_rate;
        if(i > 0){
            i = min(16, i);
        }else if(i < 0){
            i = max(-16, i);
        }
        stroker->shade_rate += i;
    }

    talking_string = stage->talking_string;

    chr = (unsigned char)talking_string[stage->talking_index_to];
    if(chr == '\0'){
        if(stage->lua_yield_state & LUA_YIELD_TALK){
            stage->lua_yield_state ^= LUA_YIELD_TALK;
            resume_flag = 1;
/*            printf("%d:talk end\n", stage->stage_num);*/
        }
    }
    if(keyhit){
        if(stage->lua_yield_state & LUA_YIELD_KEY_ENTER){
            stage->lua_yield_state ^= LUA_YIELD_KEY_ENTER;
            resume_flag = 1;
/*            printf("%d:key enter\n", stage->stage_num);*/
        }
        else{
            while(chr != '\0'){
                stage->talking_index_to += 1 + is_sjis_2byte_1st(chr);
                chr = talking_string[stage->talking_index_to];
            }
        }
    }
    else{
        if(chr != '\0'){
            stage->talking_index_to += 1 + is_sjis_2byte_1st(chr);
        }
    }
    if(!resume_flag && (stage->lua_yield_state & LUA_YIELD_FRAME)){
        stage->lua_yield_state ^= LUA_YIELD_FRAME;
        resume_flag = 1;
/*        printf("%d:wait-frame resume\n", stage->stage_num);*/
    }
    /* yield ĂȂԂłȂCxgȂ΁ACxgIB */
    if(!resume_flag && !stage->lua_yield_state){
        resume_flag = 1;
/*        printf("%d:not-inside yield resume\n", stage->stage_num);*/
    }

    /* Lua𑱍sKv鎞͑sB */
    if(resume_flag){
        all_game_status = status;
        if(lua_resume(stage->L, 0) != 0){
            /* ͂⑱słȂ̂Ȃ΃CxgIB */
            /* tF[hAEg̓XNvgŎsĂƉ肷B */
            stage->screen_fade_mode = FADEIN;
            cont = 0;
        }
        all_game_status = NULL;
    }
    return cont;
}
/*<---------------------------------------------------------------------->*/
/* bʕ`B */
static void drawscreen_event(SDL_Surface *screen, int frame, const ALL_GAME_STATUS *status)
{
    const int msg_x=64, msg_y=336, msg_w=512, msg_h=128;
    const STAGE_STATE *stage;
    const char *talking_string;
    const STROKER_STATUS *stroker;
    const TYPING_CONFIG *config;
    SDL_Rect r, dstrect, srcrect;
    int from;
    int x, y;
    int i;
    int chr;
    int chr_byte;
    char buf[3];
    SDL_Surface *surface;

    if(surface_event_bg != NULL){
        surface = surface_event_bg;
    }else{
        surface = surface_bg;
    }
    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface->w;
    dstrect.h = surface->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface->w;
    srcrect.h = surface->h;
    SDL_BlitSurface(surface, &srcrect, screen, &dstrect);

    if(status->stage.is_ending){
        stroker = &status->strokers[0];
        x = screen->w/2 - surface_stroker[stroker->chara_index][0]->w/2;
        draw_stroker_figure(screen, x, stroker->chara_index, stroker->shade_rate, 0);
    }
    else{
        for(i=0;i<PLAYER_MAX;i++){
            stroker = &status->strokers[i];
            /* GfBOXNvg̏܂ꍇAŗ댯̂Ŗh~B */
            config = &status->config;
            if(stroker->chara_index >= config->stroker_count_x * config->stroker_count_y)
                break;
            if(!i){
                x = min(64, screen->w/2 - surface_stroker[stroker->chara_index][0]->w);
            }else{
                x = max(screen->w/2, screen->w - 64 - surface_stroker[stroker->chara_index][0]->w);
            }
            draw_stroker_figure(screen, x, stroker->chara_index, stroker->shade_rate, i);
        }
        draw_stroker_name(screen, 32, msg_y - 36, status->strokers[0].chara_index);
        draw_stroker_name(screen, screen->w - 32 - surface_stroker_name[status->strokers[1].chara_index]->w, msg_y - 36,
                          status->strokers[1].chara_index);
    }

    boxColor(screen, msg_x-8, msg_y-8, msg_x+msg_w+7, msg_y+msg_h+7, 0x00000080);
    /* قǂ܂łɌ̈䎌`悪邾낤H */
    stage = &status->stage;
    talking_string = stage->talking_string;
    x = msg_x;
    y = msg_y;
    chr = (unsigned char)talking_string[stage->talking_index_to];
    if(chr == '\0'){
        r.x = msg_x + msg_w-16;
        r.y = msg_y + msg_h-16;
        r.w = 16;
        r.h = 16;
        SDL_FillRect(screen, &r, SDL_MapRGB(screen->format, 0, 0, 128));
    }

    for(from=0; from<=stage->talking_index_to; ){
        /* PɃRs[Ă`B */
        chr = (unsigned char)talking_string[from];
        if(is_sjis_2byte_1st(chr)){
            strncpy(buf, &talking_string[from], 2);
            buf[2] = '\0';
            chr_byte = 2;
            from += 2;
        }else if(chr == '\n'){
            /* s */
            x = msg_x;
            y += font_bmp.chr_h;
            from += 1;
            continue;
        }else{
            buf[0] = talking_string[from];
            buf[1] = '\0';
            chr_byte = 1;
            from += 1;
        }
        draw_string(&font_bmp, screen, x, y, buf);
        x += font_bmp.chr_w * chr_byte / 2;
        if(x + font_bmp.chr_w > msg_x + msg_w){
            x = msg_x;
            y += font_bmp.chr_h;
            if(y + font_bmp.chr_h > msg_y + msg_h){
                /* 䎌GAꂽ̂ł͂ǂȂ̂ŕ`IB */
                break;
            }
        }
    }
}
/*<---------------------------------------------------------------------->*/
/* ̓`FbNBPt[ɉ낤ɏłcc͂B */
/* miss ̓~X^CvBinput_count ͑łݏIɑŌB */
static int check_input(const SDL_keysym *keysyms, int keysym_count, const JPSTRING_DATA *jpstrdata, STROKER_STATUS *stroker, int *miss, int *input_count)
{
    int i;
    int input_char;
    int cand;
    size_t char_index;
    int roman_count = 0;
    int hit;
    ROMAN_PATTERN *romans[32];
    const char *jp_str;

    *input_count = 0;
    *miss = 0;
    for(i=0;i<keysym_count;i++){
        jp_str = jpstrdata->jp_string_kana[stroker->jp_current_line];
        /* ␔̏ꍇAVɌ擾B */
        if(roman_count == 0){
            roman_count = find_alphabet_pattern(romans, 32, &jp_str[stroker->jp_input_count]);
            if(!roman_count){
                fprintf(stderr, "[}p^[`̂Ђ炪ȂłBi%sj\n", &jp_str[stroker->jp_input_count]);
                return -1;
            }
        }
        /* ShiftȂ̓͂܂߂ĕ\\R[hɕϊB */
        input_char = util_SDLkeysym2inchar(&keysyms[i]);
        hit = 0;
        for(cand=0;cand<roman_count;cand++){
#if 0
            if(!stroker->com_stroker){
/*                printf("char=%c(%d) (%d)\n", input_char, input_char, keysyms[i].sym);*/
                printf("jpstr=%s\n", &jp_str[stroker->jp_input_count]);
                printf("cand=%s\n", romans[cand]->alphabet_pattern);
            }
#endif
            /* Ƀ`FbNAƂ΂B */
            char_index = stroker->checked_count;
            /* CfbNX͈͂𒴂Ăꍇ͈ӖȂ̂Ŕ΂B */
            if(romans[cand]->alphabet_length <= char_index){
                continue;
            }
            /* ܂ňv₾ΏۂɂB */
            if(char_index){
                if(strncmp(romans[cand]->alphabet_pattern,
                           stroker->input_stringcode,
                           char_index)){
                    continue;
                }
            }
            if(input_char == (unsigned char)romans[cand]->alphabet_pattern[char_index]){
                hit = 1;
                /* ͂ƂāAʕ\pɃobt@B */
                stroker->input_string[stroker->input_count] = util_inchar2visiblechar(input_char);
                stroker->input_count++;
                stroker->input_string[stroker->input_count] = '\0';
                if(romans[cand]->alphabet_length == char_index+1){
                    /* SvĂA{CfbNXi߂B */
                    stroker->jp_input_count += romans[cand]->jp_length;
                    if(jp_str[stroker->jp_input_count] == '\0'){
                        /* łݏIsJE^JEgAbvB */
                        stroker->jp_input_count = 0;
                        stroker->jp_current_line++;
                        /* obt@NAB */
                        *input_count += stroker->input_count;
                        stroker->input_string[0] = '\0';
                        stroker->input_count = 0;
                    }
                    /* JE^NAB */
                    stroker->checked_count = 0;
                    /* ␔NAB */
                    roman_count = 0;
                }
                else{
                    /* Ƃ肠r܂ňvĂꍇAJEgAbvB */
                    stroker->input_stringcode[stroker->checked_count] = input_char;
                    stroker->checked_count++;
                }
                /* r܂Ńp^[Gg邩񂩂B */
                break;
            }
        }
        /* ǂ̌ɂvȂꍇA~X^CvƂ݂ȂB */
        if(!hit){
            (*miss)++;
        }
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/*
 * ^CsOʏ|vC[COM̃^CsOB
 * 0,G[N-1
 */
static int check_typing(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    int hit, miss;
    int player_index;
    SDL_keysym keysyms_com[32];
    STROKER_STATUS *stroker;
    KEY_HISTORY *key_history;
    LIFE_HISTORY *life_history;
    int life_plus;
    int input_count;
    int keysym_count_tmp;
    const SDL_keysym *keysyms_tmp;

    /* ꂼ̃L[͂`FbNB */
    for(player_index=0; player_index<PLAYER_MAX; player_index++){
        stroker = &status->strokers[player_index];
        /* COMȂCOM𓮂BłȂȂ烆[U[̃L[͂𔻒B */
        if(stroker->com_stroker != NULL){
            keysym_count_tmp = comstroke_callback(stroker->com_stroker,
                &status->jpstrdata.jp_string_kana[stroker->jp_current_line][stroker->jp_input_count],
                keysyms_com, 32);
            if(keysym_count_tmp == -1)
                return -1;
            keysyms_tmp = keysyms_com;
        }
        else{
            keysym_count_tmp = keysym_count;
            keysyms_tmp = keysyms;
        }

        if(check_input(keysyms_tmp, keysym_count_tmp,
                       &status->jpstrdata, stroker, &miss, &input_count) == -1){
            return -1;
        }
        /* Ō */
        hit = keysym_count_tmp - miss;
        /* ~XŌCŌZ */
        stroker->score.miss_count += miss;
        stroker->score.input_count += hit;
        stroker->score.total_msec = status->last_tick - status->stage.stage_first_tick + 1; /* ͂܂̂+1B */
        /* Ō쐬 */
        if(hit || miss){
            key_history = &stroker->key_history[stroker->key_history_index];
            key_history->tick = status->last_tick;
            key_history->hit = hit;
            key_history->miss = miss;
            stroker->key_history_index++;
            if(stroker->key_history_index == KEY_HISTORY_MAX){
                stroker->key_history_index = 0;
            }
        }
        /* CtZ */
        life_plus = (hit - miss*4)*2;
        if(life_plus){
            stroker->life += life_plus;
            status->strokers[!player_index].life -= life_plus;
            /* Ctω쐬 */
            life_history = &stroker->life_history[stroker->life_history_index];
            life_history->tick = status->last_tick;
            life_history->life_plus = life_plus;
            stroker->life_history_index++;
            if(stroker->life_history_index == LIFE_HISTORY_MAX){
                stroker->life_history_index = 0;
            }
        }
        /* ʉBȂƂɏ͕̂ςȂB */
        if(!player_index){
            if(life_plus > 0){
                Mix_PlayChannel(-1, status->stage.mix_chunk_hit, 0);
            }else if(life_plus < 0){
                Mix_PlayChannel(-1, status->stage.mix_chunk_miss, 0);
            }
        }
        /* SłĂ܂A[vB */
        if(stroker->jp_current_line >= status->jpstrdata.jp_string_count){
            stroker->jp_current_line = 0;
        }
    }
    /* Ct\XVB */
    for(player_index=0; player_index<PLAYER_MAX; player_index++){
        stroker = &status->strokers[player_index];
        if(stroker->life_gage < stroker->life){
            stroker->life_gage++;
        }else if(stroker->life_gage > stroker->life){
            stroker->life_gage--;
        }
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* ŋ߂̃Ctωʂ߂B */
static int get_past_life(const STROKER_STATUS *stroker, Uint32 tick)
{
    int i;
    int life = 0;

    /* L^zɒHčŋ߂̃Ctωʂ߂B */
    for(i=stroker->life_history_index-1; i>=0; i--){
        if(stroker->life_history[i].tick+LIFE_HISTORY_MSEC < tick){
            return life;
        }
        life += stroker->life_history[i].life_plus;
    }
    for(i=LIFE_HISTORY_MAX-1; i>stroker->life_history_index-1; i--){
        if(stroker->life_history[i].tick+LIFE_HISTORY_MSEC < tick){
            return life;
        }
        life += stroker->life_history[i].life_plus;
    }
    /* ɗ邱ƂƂ܂BȂ͂B */
    fprintf(stderr, "xFCtωʎZopz񂪂ӂĂ鋰ꂪ܂B\n");
    return life;
}
/*<---------------------------------------------------------------------->*/
/* ŋ߂̑Ōƃ~XŌ߂B͂gĂȂ֐B */
static int get_past_key(const STROKER_STATUS *stroker, Uint32 tick, int *miss)
{
    int i;
    int hit = 0;

    *miss = 0;
    /* L^zɒHčŋ߂̃Ctωʂ߂B */
    for(i=stroker->key_history_index-1; i>=0; i--){
        if(stroker->key_history[i].tick+KEY_HISTORY_MSEC < tick){
            return hit;
        }
        *miss += stroker->key_history[i].miss;
        hit += stroker->key_history[i].hit;
    }
    for(i=LIFE_HISTORY_MAX-1; i>stroker->key_history_index-1; i--){
        if(stroker->key_history[i].tick+KEY_HISTORY_MSEC < tick){
            return hit;
        }
        *miss += stroker->key_history[i].miss;
        hit += stroker->key_history[i].hit;
    }
    /* ɗ邱ƂƂ܂BȂ͂B */
    fprintf(stderr, "xFŌZopz񂪂ӂĂ鋰ꂪ܂B\n");
    return hit;
}
/*<---------------------------------------------------------------------->*/
/* ^CsOJnB */
static int start_typing(ALL_GAME_STATUS *status)
{
    char stage_file[32];
    int player_index;
    STROKER_STATUS *stroker;

    if(status->strokers[1].com_stroker != NULL){
        sprintf(stage_file, "txt/com%02d.txt", status->strokers[1].chara_index);
        if(comstroke_load(status->strokers[1].com_stroker, stage_file) == -1){
            printf("failed to load com txt. (%s)\n", stage_file);
            return -1;
        }
    }
    /* fobOp  COM  1PCOMǂݍ */
    if(status->jpstrdata.allcom_option){
        if(status->strokers[0].com_stroker == NULL){
            status->strokers[0].com_stroker = comstroke_alloc();
        }
        comstroke_load(status->strokers[0].com_stroker, "txt/com_debug.txt");
    }
#if 1
    status->strokers[0].life = 158;
    status->strokers[1].life = 158;
#else
    status->strokers[0].life = 315;
    status->strokers[1].life = 1;
#endif
    status->strokers[0].life_gage = status->strokers[0].life;
    status->strokers[1].life_gage = status->strokers[1].life;
    for(player_index=0; player_index<PLAYER_MAX; player_index++){
        stroker = &status->strokers[player_index];
        memset(stroker->life_history, 0, sizeof stroker->life_history);
        memset(stroker->key_history, 0, sizeof stroker->key_history);
        memset(&stroker->score, 0, sizeof stroker->score);
        stroker->score.total_msec = 1;
        stroker->life_history_index = 0;
        stroker->key_history_index = 0;
        stroker->checked_count = 0;
        stroker->input_count = 0;
        stroker->input_string[0] = '\0';
        stroker->jp_input_count = 0;
    }
    /* ͕̈ʒu킹B */
    status->strokers[0].jp_current_line++;
    if(status->strokers[0].jp_current_line == status->jpstrdata.jp_string_count)
        status->strokers[0].jp_current_line = 0;
    status->strokers[1].jp_current_line = status->strokers[0].jp_current_line;
    /* 퓬JnB */
    status->stage.battle_first_tick = status->last_tick + 800*3 + 400-1;
    status->stage.battle_end = 0;
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* ^CsOʏB */
static int callback_typing(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    SDLKey sym;
    int player_index;
    int keyhit;
    int i;

    if(!status->stage.battle_end){
        if(status->stage.screen_fade_mode == FADEIN){
            /* ԍŏ̃tF[hCB */
            status->stage.screen_shade = max(0, status->stage.screen_shade - 16);
            if(!status->stage.screen_shade){
                status->stage.screen_fade_mode = FADENONE;
            }
            return 1;
        }
        if(status->stage.battle_first_tick){
            /* ^CsOJnÕJEg_EB */
            if(status->last_tick < status->stage.battle_first_tick){
                return 1;
            }
            status->stage.battle_first_tick = 0;
            /* ^CsOJňo߃~bL^B */
            status->stage.stage_first_tick = status->last_tick;
        }
    }

    /* ^CsȌB */
    if(!status->stage.battle_end){
        /* ^CsO */
        if(check_typing(keysyms, keysym_count, status) == -1)
            return -1;
        /* Ct */
        if(status->strokers[0].life_gage <= 0 ||
           status->strokers[1].life_gage <= 0){
            status->stage.battle_end = 1;
            for(player_index=0; player_index<PLAYER_MAX; player_index++){
                status->strokers[player_index].animation_count = 0;
            }
        }
    }

    /* 퓬̌oߏB */
    if(status->stage.battle_end){
        for(player_index=0; player_index<PLAYER_MAX; player_index++){
            status->strokers[player_index].animation_count++;
        }
        if(status->stage.screen_fade_mode == FADEOUT){
            status->stage.screen_shade = min(255, status->stage.screen_shade + 16);
            if(status->stage.screen_shade == 255){
                return 0;
            }
        }
        else{
            keyhit = 0;
            for(i=0;i<keysym_count;i++){
                sym = keysyms[i].sym;
                if(sym == SDLK_RETURN){
                    keyhit = 1;
                }
            }
            if(keyhit){
                status->stage.score[status->stage.stage_num] = status->strokers[0].score;
                status->stage.screen_fade_mode = FADEOUT;
                if(Mix_PlayingMusic()){
                    Mix_FadeOutMusic(500);
                }
            }
        }
    }

    return 1;
}
/*<---------------------------------------------------------------------->*/
/* Ōƃ~X̕`B */
static void draw_stroker_score(SDL_Surface *screen, int x, int y, const TYPING_SCORE *score)
{
    char str[64];

    boxColor(screen, x, y, x+160, y+16, 0x00000080);
    sprintf(str, "Hit:%3d  Miss:%3d", score->input_count, score->miss_count);
    draw_string(&font_bmp, screen, x+8, y+1, str);

    boxColor(screen, x+176, y, x+176+200, y+16, 0x00000080);
    sprintf(str, "Hit:%4.2f/s  Miss:%4.2f/s", (double)score->input_count*1000/score->total_msec,
                                              (double)score->miss_count*1000/score->total_msec);
    draw_string(&font_bmp, screen, x+184, y+1, str);
}
/*<---------------------------------------------------------------------->*/
/* Ō҂̃Ct̕ωʂ`悷B */
static void draw_stroker_lifeplus(SDL_Surface *screen, int x, int y, const STROKER_STATUS *stroker, Uint32 tick)
{
    int lifeplus;

    lifeplus = get_past_life(stroker, tick) / 2;    /* ܂蒷Ə㉺ɂ͂ݏôŊB */
    boxColor(screen, x, y-12, x+16, y+11, 0x00000080);
    if(lifeplus > 0){
        boxColor(screen, x+2, y-1-lifeplus, x+13, y-1, 0x0000ffc0);
    }
    else if(lifeplus < 0){
        boxColor(screen, x+2, y, x+13, y-lifeplus, 0xff0000c0);
    }
}
/*<---------------------------------------------------------------------->*/
/* Ō҈l̏Ԃ`悷B */
static void draw_stroker_status(SDL_Surface *screen, int x, int y, const JPSTRING_DATA *jpstrdata, const STROKER_STATUS *stroker, int roman_typing, Uint32 tick)
{
    const char *jp_str;
    char jp_str_tmp[STRING_KANA_MAX];
    int next_line;
    int dx;

    /* ̓͑Ώەij`B */
    if(stroker->jp_current_line+1 >= jpstrdata->jp_string_count){
        next_line = 0;
    }else{
        next_line = stroker->jp_current_line+1;
    }
    boxColor(screen, x+1, y+13, x+2+476, y+14+16, 0x00000080);
    draw_string(&font_bmp, screen, x+2, y+15, jpstrdata->jp_string_kanji[next_line]);

    /* ݂̓͑Ώەij`B */
    boxColor(screen, x+1, y+47, x+2+476, y+48+16, 0x00000080);
    draw_string(&font_bmp, screen, x+2, y+48, jpstrdata->jp_string_kanji[stroker->jp_current_line]);

    /* ͗p{ꕶ` */
    boxColor(screen, x+1, y+69, x+2+476, y+70+16, 0x00000080);
    jp_str = jpstrdata->jp_string_kana[stroker->jp_current_line];
    /* ͍ς݂̓{ꕶRs[B */
    strncpy(jp_str_tmp, jp_str, stroker->jp_input_count);
    jp_str_tmp[stroker->jp_input_count] = '\0';
    /* ͍ς݂̓{ꕶFtŕ`B */
    draw_string_rgba(&font_bmp, screen, x+2, y+70, jp_str_tmp, 0xff0000ff);

    /* ͂̓{ꕶ͐FȂŕ`B */
    dx = font_bmp.chr_w * (int)stroker->jp_input_count / 2;
    draw_string(&font_bmp, screen, x+2+dx, y+70, &jp_str[stroker->jp_input_count]);

    /* [}ł͓̎͂At@xbg\B */
    boxColor(screen, x+1, y+93, x+2+476, y+94+16, 0x00000080);
    if(roman_typing){
        draw_string(&font_bmp, screen, x+2, y+94, stroker->input_string);
    }
}
/*<---------------------------------------------------------------------->*/
/* winlose = 0:win 1:lose */
static void draw_winlose(SDL_Surface *screen, int x, int y, int winlose)
{
    SDL_Rect dstrect, srcrect;

    dstrect.x = x;
    dstrect.y = y;
    dstrect.w = surface_winlose->w/2;
    dstrect.h = surface_winlose->h;
    srcrect.x = winlose*surface_winlose->w/2;
    srcrect.y = 0;
    srcrect.w = surface_winlose->w/2;
    srcrect.h = surface_winlose->h;
    SDL_BlitSurface(surface_winlose, &srcrect, screen, &dstrect);
}
/*<---------------------------------------------------------------------->*/
/* ^CsOʕ`B */
static void drawscreen_typing(SDL_Surface *screen, const ALL_GAME_STATUS *status)
{
    const char *countnum = "O\0P\0Q\0R";
    const JPSTRING_DATA *jpstrdata;
    const STROKER_STATUS *strokers;
    int roman_typing;
    Uint32 stage_tick;
    Uint32 count_down;
    SDL_Rect r, dstrect, srcrect;
    int player_index;
    int count, y;
    char str[32];
    SDL_Surface *surface;

    if(surface_event_bg != NULL){
        surface = surface_event_bg;
    }else{
        surface = surface_bg;
    }
    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface->w;
    dstrect.h = surface->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface->w;
    srcrect.h = surface->h;
    SDL_BlitSurface(surface, &srcrect, screen, &dstrect);

    dstrect.w = surface_gage->w;
    dstrect.h = surface_gage->h;
    srcrect.w = surface_gage->w;
    srcrect.h = surface_gage->h;
    SDL_BlitSurface(surface_gage, &srcrect, screen, &dstrect);

    jpstrdata = &status->jpstrdata;
    strokers = status->strokers;
    roman_typing = status->config.roman_typing;

    draw_stroker_face(screen, 16, 200, strokers[0].chara_index);
    draw_stroker_status(screen, 16, 288, jpstrdata, &strokers[0], roman_typing, status->last_tick);
    draw_stroker_score(screen, 16, 268, &strokers[0].score);
    draw_stroker_lifeplus(screen, 112, 256, &strokers[0], status->last_tick);

    draw_stroker_face(screen, 560, 248, strokers[1].chara_index);
    draw_stroker_status(screen, 144, 112, jpstrdata, &strokers[1], roman_typing, status->last_tick);
    draw_stroker_score(screen, 248, 226, &strokers[1].score);
    draw_stroker_lifeplus(screen, 496, 256, &strokers[1], status->last_tick);

    /* Ct` */
    r.x = 162;
    r.y = 250;
    r.w = max(0, strokers[0].life_gage);
    r.h = 12;
    if(r.w > 0){
        boxColor(screen, r.x, r.y, r.x+r.w-1, r.y+r.h-1, 0x000080c0);
    }
    r.x += r.w;
    r.w = max(0, strokers[1].life_gage);
    if(r.w > 0){
        boxColor(screen, r.x, r.y, r.x+r.w-1, r.y+r.h-1, 0x800000c0);
    }

    if(status->stage.battle_first_tick){
        /* Xe[WJnɏo傫Ȑ̕`B */
        count_down = status->stage.battle_first_tick - status->last_tick;
        /* ̃}WbNio[̈Ӗstart_typing()QƁB */
        count_down = (count_down+400)/800;
        draw_string(&font_num96, screen, 272, 192, &countnum[count_down*3]);
    }
    else{
        /* oߎԂ̕`B */
        stage_tick = status->strokers[0].score.total_msec;
        sprintf(str, "%3d.%02d", (int)stage_tick/1000,
                                 ((int)stage_tick/10)%100);
        boxColor(screen, 292, 60, 348, 84, 0x00000080);
        draw_string(&font_bmp, screen, 296, 64, str);
    }

    if(status->stage.battle_end){
        for(player_index=0; player_index<2; player_index++){
            count = strokers[player_index].animation_count;
            if(count > 16)
                count = 16;
            y = count*count/2;
            if(strokers[player_index].life > 0){
                draw_winlose(screen, player_index*440, y, 0);
            }
            else{
                draw_winlose(screen, player_index*440, y, 1);
            }
        }
    }
}
/*<---------------------------------------------------------------------->*/
/* LO\[gp֐B̓CťvZƓB */
static int cmp_ranking(const void *p1, const void *p2)
{
    const TYPING_SCORE *sp1,*sp2;

    sp1 = &((const TYPING_RANKING *)p1)->score;
    sp2 = &((const TYPING_RANKING *)p2)->score;
    return (sp2->input_count - sp1->input_count) - (sp2->miss_count - sp1->miss_count)*4;
}
/*<---------------------------------------------------------------------->*/
/* XRAʏB */
static int callback_score(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    SDLKey sym;
    int i;
    int cont = 1;
    TYPING_RANKING *rank;
    TYPING_SCORE *score;
    STAGE_STATE *stage;
    const TYPING_SCORE *score2;
    const TYPING_CONFIG *config;

    config = &status->config;
#if 0
    /* Âꍇ̓tF[hCB */
    if(status->stage.screen_fade_mode == FADEIN){
        status->stage.screen_shade = max(0, status->stage.screen_shade - 16);
        if(!status->stage.screen_shade){
            status->stage.screen_fade_mode = FADENONE;
        }
    }
#endif
    stage = &status->stage;
    /* tF[hCEtF[hAEg */
    if(stage->screen_fade_mode == FADEIN){
        stage->screen_shade = max(0, stage->screen_shade - 16);
        if(!stage->screen_shade){
            stage->screen_fade_mode = FADENONE;
        }
    }
    else if(stage->screen_fade_mode == FADEOUT){
        stage->screen_shade = min(255, stage->screen_shade + 16);
        if(stage->screen_shade == 255){
            stage->screen_fade_mode = FADEIN;
            cont = 0;
        }
    }
    if(cont){
        for(i=0;i<keysym_count;i++){
            sym = keysyms[i].sym;
            if(sym == SDLK_RETURN){
                stage->screen_fade_mode = FADEOUT;
            }
        }
    }
    else{
        /* LOz̍Ōɍ̃XRAă\[gB */
        rank = &status->stage.typing_ranking[RANKING_MAX];
        score = &rank->score;
        memset(score, 0, sizeof(TYPING_SCORE));
        for(i=0; i<config->stroker_count_x*config->stroker_count_y; i++){
            score2 = &status->stage.score[i];
            score->input_count += score2->input_count;
            score->miss_count += score2->miss_count;
            score->total_msec += score2->total_msec;
        }
        rank->regist_time = time(NULL);
        qsort(status->stage.typing_ranking, RANKING_MAX+1, sizeof(TYPING_RANKING), cmp_ranking);
    }
    return cont;
}
/*<---------------------------------------------------------------------->*/
/* XRAʕ`B */
static void drawscreen_score(SDL_Surface *screen, const ALL_GAME_STATUS *status)
{
    SDL_Rect dstrect, srcrect;
    char str[32];
    int i,y;

    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface_bg->w;
    dstrect.h = surface_bg->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface_bg->w;
    srcrect.h = surface_bg->h;
    SDL_BlitSurface(surface_bg, &srcrect, screen, &dstrect);

    boxColor(screen, 223, 31, 416, 48, 0x00000080);
    draw_string(&font_bmp, screen, 288, 32, "SCORE");
    for(i=0; i<status->stage.stage_num; i++){
        y = 64+i*40;
        boxColor(screen, 120, y-2, 193, y+17, 0x00000080);
        sprintf(str, "Stage %2d", i+1);
        draw_string(&font_bmp, screen, 128, y, str);
        draw_stroker_score(screen, 200, y, &status->stage.score[i]);
    }
}
/*<---------------------------------------------------------------------->*/
/* LOʏB */
static int callback_ranking(const SDL_keysym *keysyms, int keysym_count, ALL_GAME_STATUS *status)
{
    SDLKey sym;
    int i;
    int cont = 1;
    STAGE_STATE *stage;

    stage = &status->stage;
    /* tF[hCEtF[hAEg */
    if(stage->screen_fade_mode == FADEIN){
        stage->screen_shade = max(0, stage->screen_shade - 16);
        if(!stage->screen_shade){
            stage->screen_fade_mode = FADENONE;
        }
    }
    else if(stage->screen_fade_mode == FADEOUT){
        stage->screen_shade = min(255, stage->screen_shade + 16);
        if(stage->screen_shade == 255){
            stage->screen_fade_mode = FADEIN;
            cont = 0;
        }
    }

    if(cont){
        for(i=0;i<keysym_count;i++){
            sym = keysyms[i].sym;
            if(sym == SDLK_RETURN){
                status->stage.screen_fade_mode = FADEOUT;
            }
        }
    }
    return cont;
}
/*<---------------------------------------------------------------------->*/
/* LOʕ`B */
static void drawscreen_ranking(SDL_Surface *screen, const ALL_GAME_STATUS *status)
{
    char str[32];
    SDL_Rect dstrect, srcrect;
    int i,y;

    dstrect.x = 0;
    dstrect.y = 0;
    dstrect.w = surface_bg->w;
    dstrect.h = surface_bg->h;
    srcrect.x = 0;
    srcrect.y = 0;
    srcrect.w = surface_bg->w;
    srcrect.h = surface_bg->h;
    SDL_BlitSurface(surface_bg, &srcrect, screen, &dstrect);

    boxColor(screen, 222, 30, 417, 49, 0x00000080);
    draw_string(&font_bmp, screen, 288, 32, "RANKING");
    for(i=0; i<RANKING_MAX; i++){
        if(!status->stage.typing_ranking[i].score.input_count)
            break;
        y = 64+i*48;
        boxColor(screen, 14, y-2, 24, y+17, 0x00000080);
        sprintf(str, "%d", i+1);
        draw_string(&font_bmp, screen, 16, y, str);
        boxColor(screen, 38, y-2, 184, y+17, 0x00000080);
        strftime(str, sizeof(str), "%y/%m/%d %H:%M:%S", localtime(&status->stage.typing_ranking[i].regist_time));
        draw_string(&font_bmp, screen, 40, y, str);
        draw_stroker_score(screen, 200, y, &status->stage.typing_ranking[i].score);
    }
}
/*<---------------------------------------------------------------------->*/
/* FPSʍɕ`B */
static void draw_fps(SDL_Surface *screen, int fps)
{
    char fps_str[16];
    SDL_Rect r;

    sprintf(fps_str, "FPS:%02d/%02d", fps, FRAME_PER_SEC);
    r.x = r.y = 0;
    r.w = font_bmp.chr_w * 5;
    r.h = font_bmp.chr_h;
    SDL_FillRect(screen, &r, SDL_MapRGB(screen->format, 0, 64, 128));
    draw_string(&font_bmp, screen, 0, 0, fps_str);
}
/*<---------------------------------------------------------------------->*/
/* fobOpB */
static void draw_debug(SDL_Surface *screen, const ALL_GAME_STATUS *status)
{
    char str[256];
    SDL_Rect r;

    sprintf(str, "state=%d", status->gamestate);
    r.x = 0;
    r.y = screen->h - font_bmp.chr_h;
    r.w = screen->w;
    r.h = font_bmp.chr_h;
    SDL_FillRect(screen, &r, SDL_MapRGB(screen->format, 0, 64, 128));
    draw_string(&font_bmp, screen, r.x, r.y, str);
}
/*<---------------------------------------------------------------------->*/
/* Sin֐le[u쐬B */
static void generate_sin_table(void)
{
    double pi;
    double rad, d;
    int i;

    pi = asin(1.0)*2.0;
    for(i=0;i<SIN_TABLE_MAX;i++){
        rad = (pi*i*2.0)/SIN_TABLE_MAX;
        d = 256*sin(rad);
        /* ΒlŎľܓB */
        if(d > 0){
            sin_table[i] = (int)(d + 0.5);
        }else{
            sin_table[i] = (int)(d - 0.5);
        }
    }
}
/*<---------------------------------------------------------------------->*/
/* 牺͌J֐B */
/* vO̎sJnɂP񂾂sB */
int act_init(void **state_data)
{
    ALL_GAME_STATUS *status;
    const char *chrimg="chr/chr2.png", *font_name=NULL;

    SDL_WM_SetCaption("dadada", "dadada");
    SDL_EnableUNICODE(1);
    srand(time(NULL));
    status = (ALL_GAME_STATUS *)malloc(sizeof(ALL_GAME_STATUS));
    memset(status, 0, sizeof(ALL_GAME_STATUS));

    load_config(&status->config);
    load_ranking(status->stage.typing_ranking, RANKING_MAX);

    /* WindowsŃtHgw肳Ă炻D悷B */
#if defined(WIN32)
    if(status->config.font_name[0] != '\0'){
        chrimg = NULL;
        font_name = status->config.font_name;
    }
#endif
    if(load_char_bmp(&font_bmp, "chr/chr2.bin", chrimg, 16, 16, font_name) == -1){
        printf("failed to load font bmp.\n");
        return -1;
    }
    if(load_char_bmp(&font_num96, "chr/num96.bin", "chr/num96.png", 96, 96, NULL) == -1){
        printf("failed to load font bmp.\n");
        return -1;
    }
    if(load_images(&status->config) == -1){
        printf("failed to load images.\n");
        return -1;
    }

    luautil_init();
    status->stage.L = luautil_LoadLua("story/story.lua");
    if(status->stage.L == NULL){
        printf("failed to load story.lua\n");
        return -1;
    }
    lua_register(status->stage.L, "api_talk", LF_talk);
    lua_register(status->stage.L, "api_figure_shade_set", LF_figure_shade_set);
    lua_register(status->stage.L, "api_set_ending", LF_set_ending);
    lua_register(status->stage.L, "api_set_usable_chara_flag", LF_set_usable_chara_flag);
    lua_register(status->stage.L, "api_bg_set", LF_bg_set);
    lua_register(status->stage.L, "api_bgm_set", LF_bgm_set);
    lua_register(status->stage.L, "api_bgm_fadeout", LF_bgm_fadeout);
    lua_register(status->stage.L, "api_fade_set", LF_fade_set);
    lua_register(status->stage.L, "api_wait_frame", LF_wait_frame);
    lua_register(status->stage.L, "api_wait_talk_end", LF_wait_talk_end);
    lua_register(status->stage.L, "api_wait_key_enter", LF_wait_key_enter);
    if(Mix_OpenAudio(44100, AUDIO_S16, 2, 4096)){
        fprintf(stderr, "Mix_OpenAudio error. %s\n", Mix_GetError());
        return -1;
    }
    status->stage.mix_chunk_hit = Mix_LoadWAV("sound/wav/reflect.wav");
    status->stage.mix_chunk_miss = Mix_LoadWAV("sound/wav/se01.wav");
    if(status->stage.mix_chunk_hit == NULL || status->stage.mix_chunk_miss == NULL){
        printf("failed to load wav. %s\n", Mix_GetError());
        return -1;
    }
    status->stage.mix_music = NULL;
    generate_sin_table();

    status->gamestate = GS_CONFIG;
    status->strokers[0].com_stroker = NULL;
#if 1
    status->strokers[1].com_stroker = comstroke_alloc();
#else
    status->strokers[1].com_stroker = NULL;
#endif
    status->fps_counter = fps_alloc(TICKS_MAX);

    *state_data = status;
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* vO̎sIɂP񂾂sB */
void act_end(void **state_data)
{
    ALL_GAME_STATUS *status;
    const TYPING_CONFIG *config;
    int i,rev;

    status = (ALL_GAME_STATUS *)*state_data;

    config = &status->config;
    for(i=0; i<config->stroker_count_x*config->stroker_count_y; i++){
        for(rev=0;rev<2;rev++){
            if(surface_stroker[i][rev]){
                SDL_FreeSurface(surface_stroker[i][rev]);
            }
            if(surface_stroker_shade[i][rev]){
                SDL_FreeSurface(surface_stroker_shade[i][rev]);
            }
        }
        if(surface_stroker_name[i]){
            SDL_FreeSurface(surface_stroker_name[i]);
        }
    }
    for(i=0; i<config->stroker_count_x*config->stroker_count_y+1; i++){
        if(surface_stroker_face[i]){
            SDL_FreeSurface(surface_stroker_face[i]);
        }
    }
    if(surface_title)
        SDL_FreeSurface(surface_title);
    if(surface_bg)
        SDL_FreeSurface(surface_bg);
    if(surface_event_bg)
        SDL_FreeSurface(surface_event_bg);
    if(surface_gage)
        SDL_FreeSurface(surface_gage);
    if(surface_winlose)
        SDL_FreeSurface(surface_winlose);

    for(i=0;i<PLAYER_MAX;i++){
        if(status->strokers[i].com_stroker != NULL){
            comstroke_free(status->strokers[i].com_stroker);
        }
    }
    fps_free(status->fps_counter);

    if(status->stage.mix_music != NULL){
        Mix_FreeMusic(status->stage.mix_music);
    }
    if(status->stage.mix_chunk_hit != NULL){
        Mix_FreeChunk(status->stage.mix_chunk_hit);
    }
    if(status->stage.mix_chunk_miss != NULL){
        Mix_FreeChunk(status->stage.mix_chunk_miss);
    }
    Mix_CloseAudio();

    /* ݒeB */
    save_config(&status->config);
    save_ranking(status->stage.typing_ranking, RANKING_MAX);

    luautil_FreeLua(status->stage.L);
    free_char_bmp(&font_bmp);
    free_char_bmp(&font_num96);
    free(status);
}
/*<---------------------------------------------------------------------->*/
/* st[Ă΂B */
int act_callback(SDL_Surface *screen, const SDL_keysym *keysyms, int keysym_count, void **state_data)
{
    ALL_GAME_STATUS *status;
    int fps;
    int cont = 1;
    int ret;
    int i;

    status = (ALL_GAME_STATUS *)*state_data;
    /* 납n܂Ƃ܂̂ŉZĩ~bL^ŏ邩jB */
    status->last_tick = SDL_GetTicks() + 10000;
    switch(status->gamestate){
    case GS_CONFIG:
        /* RtBO */
        ret = callback_config(keysyms, keysym_count, &status->config);
        if(ret == -1)
            return -1;
        if(!ret){
            if(load_jp_string("txt/txt.txt", &status->jpstrdata) == -1){
                printf("failed to load txt.\n");
                return -1;
            }
            status->stage.screen_fade_mode = FADENONE;
            if(start_title(status) == -1)
                return -1;
            status->gamestate = GS_TITLE;
        }
        drawscreen_config(screen, status->frame, &status->config);
        break;
    case GS_TITLE:
        /* ^Cg */
        ret = callback_title(keysyms, keysym_count, &status->stage);
        if(ret == -1)
            return -1;
        if(!ret){
            /* j[ɂĕB */
            if(status->stage.title_menu_index == 0){
                if(start_charaselect(status) == -1)
                    return -1;
                status->gamestate = GS_CHARASELECT;
            }
            else if(status->stage.title_menu_index == 1){
                status->gamestate = GS_RANKING;
            }
            else{
                cont = 0;
            }
        }
        drawscreen_title(screen, &status->stage, status->frame);
        break;
    case GS_CHARASELECT:
        /* LN^[ZNg */
        ret = callback_charaselect(keysyms, keysym_count, status);
        if(ret == -1)
            return -1;
        if(!ret){
            status->stage.story_num = status->strokers[0].chara_index;
            status->stage.stage_num = 0;
            status->stage.is_ending = 0;
            status->stage.screen_fade_mode = FADEIN;
            status->gamestate = GS_NEXTSTAGE;
        }
        drawscreen_charaselect(screen, status->frame, &status->config, &status->strokers[0], status->config.usable_chara_flag);
        break;
    case GS_NEXTSTAGE:
        /* Xe[Wؑ֏ */
        if(start_event(status) == -1)
            return -1;
        status->gamestate = GS_EVENT;
        break;
    case GS_EVENT:
        /* b */
        ret = callback_event(keysyms, keysym_count, status);
        if(ret == -1)
            return -1;
        if(!ret){
            if(status->stage.is_ending){
                status->gamestate = GS_SCORE;
            }
            else{
                if(start_typing(status) == -1)
                    return -1;
                status->gamestate = GS_TYPING;
            }
        }
        drawscreen_event(screen, status->frame, status);
        break;
    case GS_TYPING:
        /* ^CsO */
        ret = callback_typing(keysyms, keysym_count, status);
        if(ret == -1)
            return -1;
        if(!ret){
            if(status->strokers[0].life > 0){
                /* ꍇ */
                status->stage.stage_num++;
                status->gamestate = GS_NEXTSTAGE;
            }
            else{
                /* ꍇ */
                status->gamestate = GS_SCORE;
            }
        }
        drawscreen_typing(screen, status);
        break;
    case GS_SCORE:
        /* SXe[Wѕ\ */
        ret = callback_score(keysyms, keysym_count, status);
        if(ret == -1)
            return -1;
        if(!ret){
            status->gamestate = GS_RANKING;
        }
        drawscreen_score(screen, status);
        break;
    case GS_RANKING:
        /* LO\ */
        ret = callback_ranking(keysyms, keysym_count, status);
        if(ret == -1)
            return -1;
        if(!ret){
            if(start_title(status) == -1)
                return -1;
            status->gamestate = GS_TITLE;
        }
        drawscreen_ranking(screen, status);
        break;
    }
    status->frame++;

    for(i=0;i<keysym_count;i++){
        if(keysyms[i].sym == SDLK_F8){
            printf("R:%08x  G:%08x  B:%08x  A:%08x\n",
                    screen->format->Rmask,
                    screen->format->Gmask,
                    screen->format->Bmask,
                    screen->format->Amask
            );
        }
        else if(keysyms[i].sym == SDLK_F9){
            /* BIvV@XbvȂ */
            status->jpstrdata.noswap_option = 1;
            /* start_typing()+1̂ōŏJn邽-1ɂĂB */
            status->strokers[0].jp_current_line = -1;
        }
        else if(keysyms[i].sym == SDLK_F10){
            /* BIvV@COM */
            status->jpstrdata.allcom_option = 1;
        }
        else if(keysyms[i].sym == SDLK_F11){
            /* IvVؑ */
            if(Mix_VolumeMusic(-1)){
                /* ݂̃{[oĂāA{[ŏɂB */
                status->stage.mix_volume = Mix_VolumeMusic(-1);
                Mix_VolumeMusic(0);
            }
            else{
                /* ̃{[ɖ߂B */
                Mix_VolumeMusic(status->stage.mix_volume);
            }
        }
        else if(keysyms[i].sym == SDLK_F12){
            /* FPS\IvVؑ */
            status->fps_show ^= 1;
        }
    }

    /* ʑŜ̃tF[hB */
    if(status->stage.screen_shade){
        boxColor(screen, 0, 0, screen->w, screen->h, status->stage.screen_shade);
    }

    /* FPŠvƕ\B */
    /* Ԑ͏ʂłĂ̂ŁAŋ߂̂ߑRƂȂ */
    /* ͏ʂƕ\iȂƂ̂^B */
    fps = fps_count_fps(status->fps_counter, status->last_tick);
    if(status->fps_show){
        draw_fps(screen, fps);
    }
#if 0
    draw_debug(screen, status);
#endif
    SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);

    return cont;
}
