#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stroke.h"
#include "jpstr.h"

#define ROMAN_PATTERN_MAX 2048

#define IS_SJIS_2BYTE(p) (is_sjis_2byte_1st((unsigned char)(p[0])) && is_sjis_2byte_2nd((unsigned char)(p[1])))

static ROMAN_PATTERN roman_patterns[ROMAN_PATTERN_MAX];
static int roman_pattern_count = 0;

/*<---------------------------------------------------------------------->*/
/* 񂩂w肵TĂ̈ʒuԂB
 * 2Byteyщ~L̎͌̕ΏۂƂȂB
 */
static const char* strchr_with_escape(const char *str, int chr)
{
    const char *p;

    p = str;
    while(*p != '\0'){
        if(is_sjis_2byte_1st((unsigned char)(*p)) &&
           is_sjis_2byte_2nd((unsigned char)(*(p+1)))){
            p++;
            if(*p == '\0')
                return NULL;
            continue;
        }
        if(*p == '\\'){
            p++;
            if(*p == '\0')
                return NULL;
        }
        else{
            if(*p == chr){
                return p;
            }
        }
        p++;
    }
    return NULL;
}
/*<---------------------------------------------------------------------->*/
/* {p^[ŏ\[g邽߂̔r֐B */
static int cmp_roman_pattern(const void *p1, const void *p2)
{
    const ROMAN_PATTERN *rp1,*rp2;

    rp1 = (const ROMAN_PATTERN *)p1;
    rp2 = (const ROMAN_PATTERN *)p2;
    return strcmp(rp1->jp_pattern, rp2->jp_pattern);
}
/*<---------------------------------------------------------------------->*/
/* {p^[̍ŏ̂P(2byte)r邽߂̔r֐B */
static int cmp_roman_pattern_first_2byte_1(const void *p1, const void *p2)
{
    const ROMAN_PATTERN *rp1,*rp2;

    rp1 = (const ROMAN_PATTERN *)p1;
    rp2 = (const ROMAN_PATTERN *)p2;
    return memcmp(rp1->jp_pattern, rp2->jp_pattern, 2);
}
/*<---------------------------------------------------------------------->*/
/* {p^[̍ŏ̎w蕶(2byte*n)r邽߂̔r֐B */
static int cmp_roman_pattern_first_2byte_n(const void *p1, const void *p2, int n)
{
    const ROMAN_PATTERN *rp1,*rp2;

    rp1 = (const ROMAN_PATTERN *)p1;
    rp2 = (const ROMAN_PATTERN *)p2;
    return memcmp(rp1->jp_pattern, rp2->jp_pattern, n*2);
}
/*<---------------------------------------------------------------------->*/
/* {p^[̍ŏ̂P(1byte)r邽߂̔r֐B */
static int cmp_roman_pattern_first_1byte_1(const void *p1, const void *p2)
{
    const ROMAN_PATTERN *rp1,*rp2;

    rp1 = (const ROMAN_PATTERN *)p1;
    rp2 = (const ROMAN_PATTERN *)p2;
    return memcmp(rp1->jp_pattern, rp2->jp_pattern, 1);
}
/*<---------------------------------------------------------------------->*/
/* {p^[̍ŏ̎w蕶(1byte*n)r邽߂̔r֐B */
static int cmp_roman_pattern_first_1byte_n(const void *p1, const void *p2, int n)
{
    const ROMAN_PATTERN *rp1,*rp2;

    rp1 = (const ROMAN_PATTERN *)p1;
    rp2 = (const ROMAN_PATTERN *)p2;
    return memcmp(rp1->jp_pattern, rp2->jp_pattern, n);
}
/*<---------------------------------------------------------------------->*/
static int add_roman_charset(const char *alstr, const char *jpstr)
{
    ROMAN_PATTERN *rp;
    if(roman_pattern_count >= ROMAN_PATTERN_MAX){
        fprintf(stderr, "[}`Ggӂ܂B\n");
        return -1;
    }
    rp = &roman_patterns[roman_pattern_count];
    strcpy(rp->alphabet_pattern, alstr);
    strcpy(rp->jp_pattern, jpstr);
    rp->alphabet_length = strlen(rp->alphabet_pattern);
    rp->jp_length = strlen(rp->jp_pattern);
    if(IS_SJIS_2BYTE(rp->jp_pattern)){
        rp->wc_count = 2;
    }else{
        rp->wc_count = 1;
    }
/*    printf("%s,%s\n", rp->jp_pattern, rp->alphabet_pattern);*/
    roman_pattern_count++;
    return 0;
}
/*<---------------------------------------------------------------------->*/
/*
 * o^ꂽp^[ŁAOɂPڂƓtp^[ꍇ͐Vɓo^B
 * ""͏o悤ɂ邽߁B
 * F",RA"->",RRA"
 */
static int generate_ltu(void)
{
    int pat_count;
    int i,j;
    ROMAN_PATTERN *rp;
    char jp_buf[JP_MAX+1];
    char alpha_buf[ALPHA_MAX+1];

    pat_count = roman_pattern_count;
    for(i=0;i<pat_count;i++){
        rp = &roman_patterns[i];
        /* ꉹ͂ɂĂ͏ȂB */
        if(strchr("AIUEO", rp->alphabet_pattern[0]) != NULL){
            continue;
        }
        sprintf(alpha_buf, "%c%s", rp->alphabet_pattern[0], rp->alphabet_pattern);
        /* `TBdȂB */
        for(j=0;j<pat_count;j++){
            if(!strcmp(roman_patterns[i].alphabet_pattern, rp->alphabet_pattern))
                break;
        }
        if(j < pat_count){
            sprintf(jp_buf, "%s", rp->jp_pattern);
            if(add_roman_charset(alpha_buf, jp_buf) == -1)
                return -1;
        }
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/*
 * o^ꂽp^[ŁAO'N'tp^[ꍇ͐Vɓo^B
 * ""Nłœ͏oꍇ͂̂悤ɏ邽߁B
 * F",RA"->",NRA"
 */
static int generate_nn(void)
{
    int pat_count;
    int i,j;
    ROMAN_PATTERN *rp;
    char jp_buf[JP_MAX+1];
    char alpha_buf[ALPHA_MAX+1];

    pat_count = roman_pattern_count;
    for(i=0;i<pat_count;i++){
        rp = &roman_patterns[i];
        /* ꉹ͂ɂĂ͏ȂB */
        if(strchr("AIUEO", rp->alphabet_pattern[0]) != NULL){
            continue;
        }
        sprintf(alpha_buf, "N%s", rp->alphabet_pattern);
        /* `TBdȂB */
        for(j=0;j<pat_count;j++){
            if(!strcmp(roman_patterns[i].alphabet_pattern, rp->alphabet_pattern))
                break;
        }
        if(j < pat_count){
            sprintf(jp_buf, "%s", rp->jp_pattern);
            if(add_roman_charset(alpha_buf, jp_buf) == -1)
                return -1;
        }
    }
    return 0;
}
/*<---------------------------------------------------------------------->*/
/* ϊ`t@CǂݍށBis_roman = 0 ̎"",""̏sȂB  */
int load_roman_charset(const char *file, int is_roman)
{
    FILE *fp;
    char buf[256];
    char jp_buf[JP_MAX+1];
    char alpha_buf[ALPHA_MAX+1];
    char *p;
    const char *p1,*p2;
    int count; 
    int line_index = 0;

    roman_pattern_count = 0;

    fp = fopen(file, "r");
    if(fp == NULL){
        return -1;
    }

    while(fgets(buf, sizeof buf, fp) != NULL){
        line_index++;
        p = strchr(buf, '\n');
        if(p != NULL){
            *p = '\0';
        }
        if(buf[0] == '\0'){
            /* s͖ */
            continue;
        }
        p1 = &buf[0];
        p2 = strchr_with_escape(buf, ',');
        if(p2 == NULL){
            printf("xFJ}ЂƂȂ񂪂܂B܂Bi%sF%dj\n", file, line_index);
            continue;
        }
        /* {p^[Ƃ΂B */
        if(p1+JP_MAX < p2){
            printf("xF{ꂪ񂪂܂B܂Bi%sF%dj\n", file, line_index);
            continue;
        }
        /* {p^[ꎞ̈ɃRs[B */
        count = 0;
        while(p1 < p2){
            jp_buf[count] = *p1;
            p1++;
            count++;
        }
        jp_buf[count] = '\0';
#if 0
        /* o[Wł2BytêݑΏۂƂĂAς1Byte悤ɂB */
        if(!is_sjis_2byte_string(jp_buf)){
            fprintf(stderr, "xFSJIS 2byteׂꏊɂłȂ܂Bi%sFLine %dF%sj\n", file, line_index, jp_buf);
            continue;
        }
#endif
        /*
         * J}ŋ؂ꂽAt@xbg𒲂ׁA
         * ̐p^[Gg𐶐B
         */
        for(;;){
            p1 = p2+1;
            if(*p1 == '\0')
                break;
            p2 = strchr_with_escape(p1, ',');
            if(p2 == NULL){
                p2 = strchr(p1, '\0');
            }
            /* At@xbgp^[ꎞ̈ɃRs[B */
            count = 0;
            while(p1 < p2){
                /* At@xbgp^[Ƃ΂B */
                if(count >= ALPHA_MAX){
                    count = 0;
                    break;
                }
                if(*p1 == '_'){
                    /* A_[o[tĂ镶̓VtgtƂčŏʃrbg𗧂ĂB */
                    if(count > 0){
                        alpha_buf[count-1] |= 0x80;
                    }
                }
                else if(*p1 == '\\'){
                    /* ~L炻̎̂̕܂܃Rs[B */
                    p1++;
                    if(*p1 == '\0')
                        break;
                    alpha_buf[count] = *p1;
                    count++;
                }
                else if(*p1 == '*'){
                    /* ͂`ĂȊO̔Cӂ̃L[ɃqbgB */
                    alpha_buf[count] = 255;
                    count++;
                }
                else{
                    /* PRs[B */
                    alpha_buf[count] = *p1;
                    count++;
                }
                p1++;
            }
            /* 󕶎Ȃo^ȂB */
            if(count > 0){
                alpha_buf[count] = '\0';
                if(add_roman_charset(alpha_buf, jp_buf) == -1){
                    /* ꂾƊÕ[vȂccB */
                    break;
                }
            }
        }
    }
    fclose(fp);

    if(is_roman){
        generate_ltu();
        generate_nn();
    }
    /* \̂\[gĂ邪A|C^̔zʂɍă\[gB */
    qsort(roman_patterns, roman_pattern_count, sizeof(ROMAN_PATTERN), cmp_roman_pattern);
#if 0
    for(count=0;count<roman_pattern_count;count++){
        printf("%4d,%s,%s\n", count, roman_patterns[count].jp_pattern, roman_patterns[count].alphabet_pattern);
    }
    printf("roman pattern size = %d\n", sizeof roman_patterns);
#endif
    return 0;
}
/*<---------------------------------------------------------------------->*/
/*
 * jp_str Ŏw肳ꂽ{̍ŏ̂PGgA
 * ΉAt@xbgp^[ւ̃|C^zɓĕԂB
 * ߂l͔zɊi[ꂽ|C^̐B
 */
int find_alphabet_pattern(ROMAN_PATTERN **romans, int alstr_max, const char *jp_str)
{
    ROMAN_PATTERN key;
    ROMAN_PATTERN *locate;
    void *p;
    int count=0;
    size_t chr_byte;
    int (*cmp_func_1)(const void *, const void *); /* r֐ւ̃|C^ */
    int (*cmp_func_n)(const void *, const void *, int);

    /* 2byter邩1byter邩fB */
    if(is_sjis_2byte_1st(*jp_str)){
        cmp_func_1 = cmp_roman_pattern_first_2byte_1;
        cmp_func_n = cmp_roman_pattern_first_2byte_n;
        chr_byte = 2;
    }else{
        cmp_func_1 = cmp_roman_pattern_first_1byte_1;
        cmp_func_n = cmp_roman_pattern_first_1byte_n;
        chr_byte = 1;
    }

    /* L[ăoCiT[`B */
    /* C++ STLlower_bound()gˁB */
    strncpy(key.jp_pattern, jp_str, JP_MAX);
    key.jp_pattern[JP_MAX] = '\0';
    p = bsearch(&key, roman_patterns, roman_pattern_count, sizeof(ROMAN_PATTERN), cmp_func_1);
    if(p == NULL)
        return 0;

    /*
     * {̍ŏ̂PGgłƂ̓rwĂ
     * Ȃ̂ŁAGgԌɖ߂B
     */
    locate = (ROMAN_PATTERN*)p;
    while(roman_patterns < locate && !cmp_func_1(locate-1, locate)){
        locate--;
    }

    /* {̃Ggւ̃|C^zɓĕԂB */
    /* \[gĂ邱ƂlƂ悭ł邾낤ǁccB */
    for(;;){
        if(count >= alstr_max){
            fprintf(stderr, "xF[}ϊp^[̌zɎ߂܂Bi%sj\n", jp_str);
            break;
        }
        /* P̃p^[AQڈȍ~p^[ΏۂƂB */
        if(locate->jp_length == chr_byte ||
           !cmp_func_n(locate, &key, locate->wc_count)){
            romans[count] = locate;
            count++;
        }
        /* z񒆂̃Gg̏IɒB邩ÃGgقȂ镨ȂIB */
        /* ẑЂƂw|C^𐶐̂ANSI C̋Ki㐳(^^; */
        if(locate+1 == roman_patterns+roman_pattern_count ||
           cmp_func_1(locate+1, locate)){
            break;
        }
        locate++;
    }
    return count;
}
